from typing import Annotated

from crypteia import Multibasing, Multihashing, ToBytes, compose
from pydantic import Field

get_hash = compose(ToBytes(), Multihashing("sha3-224"), Multibasing("base58btc"))

# sha3-224 is a 28-byte hash, which is 42 characters in base58btc
# the sha3-224 identifier (0x17) is 5d when encoded in base58btc
# base58btc encoding prepends a z to the string

HashDigest = Annotated[str, Field(pattern=r"z5d[1-9A-HJ-NP-Za-km-z]{39}")]



from pymongo.errors import PyMongoError


class RedBabyError(PyMongoError):
    """
    Generic error.
    """


class DocumentNotFound(RedBabyError):
    """
    Raised when no document is found.
    """



from typing import Any

from bson import ObjectId
from pydantic import GetCoreSchemaHandler
from pydantic.json_schema import JsonSchemaValue
from pydantic_core import core_schema


class PyObjectId(ObjectId):
    @classmethod
    def __get_pydantic_core_schema__(
        cls, _source_type: Any, _handler: GetCoreSchemaHandler
    ) -> core_schema.CoreSchema:
        # left to right evaluation on the schemas below
        from_json_validation = core_schema.chain_schema(
            [
                core_schema.str_schema(),
                core_schema.no_info_plain_validator_function(cls.validate),
            ]
        )
        from_python_validation = core_schema.union_schema(
            [core_schema.is_instance_schema(ObjectId), from_json_validation]
        )
        # we only want to stringify the ObjectId when serializing to json
        serializer = core_schema.plain_serializer_function_ser_schema(
            str, when_used="json"
        )
        pydantic_core_schema = core_schema.json_or_python_schema(
            json_schema=from_json_validation,
            python_schema=from_python_validation,
            serialization=serializer,
        )
        return pydantic_core_schema

    @classmethod
    def validate(cls, value) -> ObjectId:
        if not ObjectId.is_valid(value):
            raise ValueError(f"'{value}' is not a valid bson.ObjectId.")

        return ObjectId(value)

    @classmethod
    def __get_pydantic_json_schema__(cls, _core_schema, handler) -> JsonSchemaValue:
        return handler(core_schema.str_schema())



import logging
from typing import Any, Optional

from pydantic import BaseModel
from pymongo import IndexModel
from pymongo.collection import Collection

from ..database import DB


class BaseDocument(BaseModel):

    def bson(self) -> dict[str, Any]:
        obj = self.model_dump(by_alias=True, mode="python")
        return obj

    @classmethod
    def collection_name(cls) -> str:
        raise NotImplementedError

    @classmethod
    def collection(
        cls,
        db_name: Optional[str] = None,
        suffix: Optional[str] = None,
        alias: str = "default",
    ) -> Collection:
        return DB.get(db_name, suffix, alias)[cls.collection_name()]

    @classmethod
    def indexes(cls) -> list[IndexModel]:
        return []

    @classmethod
    def create_indexes(cls, alias: str = "default"):
        idx = cls.indexes()
        if not idx:
            logging.warning(f"No indexes for '{cls.collection_name()}'.")
            return
        cls.collection(alias=alias).create_indexes(idx)
